import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

from demv.helpers import sample


class DEMV(TransformerMixin, BaseEstimator):
    """
    Debiaser for Multiple Variables(DEMV) is a pre-processing algorithm for binary
    and multi-class datasets that mitigates bias by perfectly balancing the sensitive groups
    identified by each possible sensitive variables' value and each label's value [1].

    References:
         [1] G. d'Aloisio, A. D'Angelo, A. Di Marco, e G. Stilo,
         «Debiaser for Multiple Variables to enhance fairness in classification tasks»,
         Information Processing & Management, vol. 60, mar. 2023, doi: 10.1016/j.ipm.2022.103226.

    Based on the code from: https://github.com/giordanoDaloisio/demv
    """

    def __init__(self, sensitive_vars, round_level=1, stop=10000, verbose=False):
        """
        Parameters
        ----------
        sensitive_vars : list
            List of sensitive variable names
        round_level : float, optional
            Tolerance value to balance the sensitive groups (default is 1)
        stop : int, optional
            Maximum number of iterations to balance the sensitive groups (default is 10000)
        verbose : bool, optional
            Prints w_exp/w_obs, useful for debugging (default is False)
        """
        self.sensitive_vars = sensitive_vars
        self.disparities = []
        self.round_level = round_level
        self.stop = stop
        self.debug = verbose
        self.iter = 0
        self.label = "y"

    def fit(self, x: pd.DataFrame, y: np.ndarray):
        """
        Balances the dataset's sensitive groups

        Args
        ----------
        x : pd.DataFrame
            Dataset to be balanced
        y : array-like
            Labels of the dataset

        Returns
        -------
         x: Balanced dataset
         y: Balanced labels of the dataset
        """
        return self.fit_transform(x, y)

    def transform(self, x: pd.DataFrame, y: np.ndarray):
        """
        Balances the dataset's sensitive groups

        Args
        ----------
        x : pd.DataFrame
            Dataset to be balanced
        y : array-like
            Labels of the dataset

        Returns
        -------
         x: Balanced dataset
         y: Balanced labels of the dataset
        """
        return self.fit_transform(x, y)

    def fit_transform(self, x, y):
        """
        Balances the dataset's sensitive groups

        Args
        ----------
        x : pd.DataFrame
            Dataset to be balanced. The sensitive variables must be the index of the dataframe
        y : array-like
            Labels of the dataset

        Returns
        -------
         x: Balanced dataset
         y: Balanced labels of the dataset
        """
        x = x.copy()
        x[self.label] = y
        x_new, disparities, iters = sample(
            x,
            self.sensitive_vars,
            self.label,
            self.round_level,
            self.debug,
            0,
            [],
            True,
            self.stop,
        )
        self.disparities = disparities
        self.iter = iters
        y_new = x_new[self.label].values
        x_new = x_new.drop(self.label, axis=1)
        return x_new, y_new

    def get_iters(self):
        """
        Gets the maximum number of iterations

        Returns:
        int: maximum number of iterations
        """
        return self.iter

    def get_disparities(self):
        """
        Returns the list of w_exp/w_obs

        Returns:
        list: list of disparities values
        """
        return self.disparities



import pandas as pd
import numpy as np


def _balance_set(
    w_exp, w_obs, df: pd.DataFrame, tot_df, round_level=None, debug=False, k=-1
):
    if w_obs == 0:
        return df, 0, 0
    disp = round(w_exp / w_obs, round_level) if round_level else w_exp / w_obs
    disparity = [disp]
    i = 0
    while disp != 1 and i != k and w_obs != 0:
        if w_exp / w_obs > 1:
            df = pd.concat([df, df.sample()])
        elif w_exp / w_obs < 1:
            df = df.drop(df.sample().index, axis=0)
        w_obs = len(df) / len(tot_df)
        if w_obs != 0:
            disp = round(w_exp / w_obs, round_level) if round_level else w_exp / w_obs
            disparity.append(disp)
            if debug:
                print(w_exp / w_obs)
            i += 1
        else:
            return df, disparity, i
    if i == k:
        print("Warning: max iterations reached")
    return df, disparity, i


def sample(
    d: pd.DataFrame,
    s_vars: list,
    label: str,
    round_level: float,
    debug: bool = False,
    i: int = 0,
    G=None,
    cond: bool = True,
    stop: int = 10000,
):
    if G is None:
        G = []
    d = d.copy()
    n = len(s_vars)
    disparities = []
    iter = 0
    if i == n:
        for l in np.unique(d[label]):
            g = d[(cond) & (d[label] == l)]
            if len(g) > 0:
                w_exp = (len(d[cond]) / len(d)) * (len(d[d[label] == l]) / len(d))
                w_obs = len(g) / len(d)
                g_new, disp, k = _balance_set(
                    w_exp, w_obs, g, d, round_level, debug, stop
                )
                g_new = g_new.astype(g.dtypes.to_dict())
                disparities.append(disp)
                G.append(g_new)
                iter = max(iter, k)
        return G, iter, disparities
    else:
        s = s_vars[i]
        i = i + 1
        G1, k1, disp1 = sample(
            d,
            s_vars,
            label,
            round_level,
            debug,
            i,
            G.copy(),
            cond=cond & (d[s] == 0),
            stop=stop,
        )
        G2, k2, disp2 = sample(
            d,
            s_vars,
            label,
            round_level,
            debug,
            i,
            G.copy(),
            cond=cond & (d[s] == 1),
            stop=stop,
        )
        G += G1
        G += G2
        iter = max([iter, k1, k2])
        new_disps = disp1 + disp2
        disparities.append(new_disps)
        limit = 1
        for s in s_vars:
            limit *= len(np.unique(d[s]))
        if len(G) == limit * len(np.unique(d[label])):
            return (
                pd.DataFrame(
                    G.pop().append([g for g in G]).sample(frac=1, random_state=2)
                ),
                disparities,
                iter,
            )
        else:
            return G, iter, disparities



import numpy as np
import pandas as pd
from demv import DEMV
import unittest


class TestDEMV(unittest.TestCase):
    def setUp(self):
        # Define some sample data for testing
        np.random.seed(42)
        num_samples = 100
        self.x = pd.DataFrame(
            {
                "feature1": np.random.rand(num_samples),
                "feature2": np.random.rand(num_samples),
                "sensitive_var1": np.random.randint(0, 2, num_samples),
                "sensitive_var2": np.random.randint(0, 2, num_samples),
            }
        )
        self.y = np.random.randint(0, 2, num_samples)

    def test_fit_transform(self):
        # Create an instance of the DEMV class
        demv = DEMV(sensitive_vars=["sensitive_var1", "sensitive_var2"])

        # Test the fit_transform method
        x_balanced, y_balanced = demv.fit_transform(self.x, self.y)

        # Check if the output has the correct shape
        self.assertEqual(x_balanced.shape[0], len(y_balanced))

        # Check if the output has the correct type
        self.assertIsInstance(x_balanced, pd.DataFrame)

        # Check if the output has the correct type
        self.assertIsInstance(y_balanced, np.ndarray)

    def test_get_iters(self):
        # Create an instance of the DEMV class
        demv = DEMV(sensitive_vars=["sensitive_var1", "sensitive_var2"])

        # Test the get_iters method before fit_transform
        self.assertEqual(demv.get_iters(), 0)

        # Test the get_iters method after fit_transform
        _, _ = demv.fit_transform(self.x, self.y)
        self.assertNotEqual(demv.get_iters(), 0)

    def test_get_disparities(self):
        # Create an instance of the DEMV class
        demv = DEMV(sensitive_vars=["sensitive_var1", "sensitive_var2"])

        # Test that the get_disparities method returns an empty array before fit transform
        disparities = demv.get_disparities()
        self.assertEqual(disparities, [])

        # Test that the get_disparities method returns a non-empty array after fit transform
        demv.fit_transform(self.x, self.y)
        disparities = demv.get_disparities()
        self.assertNotEqual(disparities, [])


if __name__ == "__main__":
    unittest.main()
